################################################################################
# insert-animacy.py                                                            #
#          purpose: Insert animacy information for nouns within PPs            #
#           author: Michael Percillier                                         #
#             date: 28 June 2019                                               #
#              log: 28 June 2019 --> v1                                        #
#                   15 September 2020: - Added disclaimer informer the user    #
#                                        about the use of the script in the    #
#                                        output file                           #
#                                      - Added a category "panimate"           #
#                                      --> v2                                  #
#                   6 October 2021: Updated BASICS Toolkit in the disclaimer   #
#                                   --> v3                                     #
################################################################################

import codecs
import re
import sre_constants # Handle Regex errors
import argparse
import os
import sys

# Set version number
versionNumber = 3

# Define disclaimer
disclaimer = """
/*
DISCLAIMER:
This corpus file was treated by a script to add information about the animacy status of nouns occurring in prepositional phrases headed by "to" in the vicinity of verbs to help determine the status of prepositional phrases as prepositional objects or adverbials: insert-animacy.py, version %d, created by Michael Percillier within the research project Borrowing of Argument Structure in Contact Situations (BASICS).

Each piece of inserted information is demarcated by @ characters and specified by an attribute, such as:
- l: the lemma in its Present-Day English form and/or as specified in the (Parsed) Linguistic Atlas of Early Middle English, whereby multiple entries are separated by a pipe (|) character
- a: the animacy status, which may be "animate", "inanimate", "ambiguous" for homonyms that could not be distinguished, and "panimate" for lexemes that are inanimate in their primary sense but have a high propensity to refer to animate entities via metonymy or metaphor. Such cases can be included in a query for animate entities with: *@a=animate@*|*@a=panimate@*

For more information on the script "insert-animacy.py", please consult the page "Lemmatizer" on the BASICS Toolkit website: http://basics-toolkit.spdns.org/

For more information on the BASICS research project, please consult the pages https://tinyurl.com/dfgbasics and https://gepris.dfg.de/gepris/projekt/265711632
*/

""" % versionNumber

# save ANSI escape sequences to make messages to user pop out
ANSIreverse = "\033[7m"
ANSIreset = "\033[0m"

# Info banner
print "\n" + ANSIreverse
print "{:^80}".format("insert-animacy.py: Insert animacy information for nouns within PPs")
print "{:^80}".format("in PPCME2, PCMEP, or PLAEME")
print "{:^80}".format("Created by Michael Percillier, University of Mannheim, 2019-2021")
print "{:^80}".format("Version %d" % versionNumber)
print ANSIreset + "\n"


# Modus operandi
# 1. Load table of noun forms with animacy information (and partial lemmas info)
# 2. In the corpus files, check every noun form and insert information when
#    applicable

# 1. Load table of noun forms and create a dictionary from it
animacyHash = {}
print "Loading annotation data..."
with codecs.open("toPP-animacy.csv", encoding="utf-8", mode="r") as inp:
    for line in inp:
        # Ignore table header
        if re.search("Form;Lemma", line):
            pass
        # All remaining lines are regular table rows
        else:
            splitLine = line.strip().split(";")
            # Indices are now as follows:
            # 0: form, 1: lemma/lexel, 2: animacy, 3: comment
            # Store these values in the dictionary
            animacyHash[splitLine[0].lower()] = {}
            animacyHash[splitLine[0].lower()]["lemma"] = splitLine[1]
            animacyHash[splitLine[0].lower()]["animacy"] = splitLine[2]
            animacyHash[splitLine[0].lower()]["comment"] = splitLine[3]
print "Annotation data loaded.\n"

# 2. Load files from selected corpus, and annotate

## 2.1. Get name of corpus directory from program call (as well as whether
##          the corpus is PLAEME, which has a different format)
try:
    parser = argparse.ArgumentParser()
    parser.add_argument("directory",help="name of the corpus directory")
    parser.add_argument(
        "-p", "--plaeme",
        help="Indicate whether PLAEME texts are to be processed",
        action="store_true")
    args = parser.parse_args()
except SystemExit:
    print "\nYou haven't specified a corpus folder. Please try again.\n"

# 2.2 Open directory containing corpus and create an output directory
fileList = os.listdir(os.getcwd() + "/" +  args.directory)
try:
    os.mkdir(os.getcwd() + "/" +  args.directory + "_Animacy")
except OSError:
    print "\nIntended output folder already exists.\nDo you wish to proceed?"
    overwriteAnswer = ""
    while not re.match("[yYnN]",overwriteAnswer):
        overwriteAnswer = raw_input("(Y)es [overwrite]\t(N)o [quit] ")
    if re.match("[nN]",overwriteAnswer):
        print "Quitting program."
        print "Delete or rename the folder *_Lemmatised and rerun the script."
        raise SystemExit

# 2.4 Define functions needed for the lemmatisation

# define function to pre-process PLAEME forms
def plaemeProcessing(form):
    #processedForm = re.sub(r"\+c","c",form) # remove any c-ligature marking
    #processedForm = re.sub("(g|G)2","\\1",processedForm) # replace alternative 'yogh' with regular 'yogh'
    #processedForm = re.sub(r"\+(w|W)","\\1",processedForm) # replace 'wynn' with 'w'
    #processedForm = re.sub("'","",processedForm) # remove accents
    #processedForm = re.sub("=","",processedForm) # remove morpheme boundary
    #processedForm = re.sub("@","",processedForm) # remove word split marker
    #processedForm = re.sub(r"\^","",processedForm) # remove superscript marker
    processedForm = re.sub(r"^(.*)\-.*?$", "\\1",form) # remove lexel
    return processedForm

# define function to check matching lemmas to a verb form
def checkLemmas(form):
    lowerForm = form.lower() # use a lowercase version of the observed form,
    # which may be in all-caps or at the beginning of a sentence
    lemmaMatches = [] # list to store the lemmas matching the form
    if lowerForm in animacyHash.keys():
        formMatches = animacyHash[lowerForm]
    elif stemmer(lowerForm) in animacyHash.keys():
        formMatches = animacyHash[stemmer(lowerForm)]
    else:
        formMatches = {}
        # Diagnostic, comment out if not needed:
        # print "No matches found for: %s" %lowerForm
    return formMatches

# define function to stem plural forms
def stemmer(form):
    stemmedForm = re.sub("[eiy]?[sz]$", "", form)
    return stemmedForm

# define function to handle matching lemmas
def handleMatches(form, formMatches):
    # the form will be replaced by itself followed by lemma attributes
    formInfo = form
    # append info to form
    formInfo = formInfo + "@l=%s@a=%s@" % (
        formMatches["lemma"], formMatches["animacy"])
    # append comment if present
    if formMatches["comment"] != "":
        formInfo = formInfo + "c=%s@" % formMatches["comment"]
    return formInfo

# 2.4 Go through each corpus file and insert animacy information to nouns
fileCounter = 0
for corpusFile in fileList:
    # Only treat corpus files, which end in *.psd
    if re.search("\.psd$",corpusFile):
        # Create output file
        outpFileName = re.sub("\.psd",".animacy.psd",corpusFile)
        outpFilePath = (os.getcwd()
            + "/" +  args.directory + "_Animacy"+"/" + outpFileName)
        outp = codecs.open(outpFilePath,encoding="utf-8",mode="w")
        # Insert disclaimer
        outp.write(disclaimer)
        inp = (codecs.open(os.getcwd() + "/" + args.directory + "/"
            + corpusFile,encoding="utf-8",mode="r"))
        # Now treat each line of the input file
        for line in inp:
            # Check if line contains an N-tagged noun:
            if re.search("\(N",line):
                lineNouns = re.findall("\(N\S* \$?[^\(\)]+",line)
                updatedLine = line # make a copy of the line to be modified
                for noun in lineNouns:# handle the noun(s) in the line
                    # handle potential split forms
                    # get tag with index digits
                    splitIndex = re.findall("\(N\w+\d+",noun)
                    try:
                        splitIndex = splitIndex[0]
                        splitIndex = re.sub("\D","",splitIndex) # keep digits only
                    except IndexError:
                        splitIndex = "00" # special for unsplit forms
                    # deal with intermediate items first
                    if splitIndex[0] != splitIndex[1]:
                        if splitIndex[1] == "1":
                            lookUpForm = re.sub("\(N\S* \$?","",noun)
                        else:
                            lookUpForm = lookUpForm + re.sub(
                                "\(N\S* \$?", "", noun)
                    # for final items of split forms and unsplit forms,
                    # a matching lemma is queried
                    else:
                        # if the noun form is not split, nounForm and
                        # lookUpForm are identical
                        if splitIndex == "00":
                            nounForm = re.sub("\(N\S* \$?","",noun)
                            lookUpForm = nounForm
                        else: # handle last item of split noun
                            # form after which the lemma info will be placed
                            nounForm = re.sub("\(N\S* \$?","", noun)
                            # form used to check for lemmas
                            lookUpForm = lookUpForm + nounForm
                    # pre-processing needed for PLAEME forms
                    if args.plaeme:
                        lookUpForm = plaemeProcessing(lookUpForm)
                    # check form for matches
                    plainMatches = checkLemmas(lookUpForm)
                    if len(plainMatches.keys()) > 0:
                        formInfo = handleMatches(nounForm,plainMatches)
                        # update line
                        # replace any round brackets by square brackets
                        formInfo = re.sub("\(", "[", formInfo)
                        formInfo = re.sub("\)", "]", formInfo)
                        # remove any whitespace that may have been introduced
                        formInfo = re.sub("\s+", "", formInfo)
                        # any + in the form need to be escaped to update
                        # the line
                        escapedNounForm = re.sub("\+","\+", nounForm)
                        try:
                            updatedLine = re.sub(
                                escapedNounForm, formInfo, updatedLine)
                        # error raised in case of traces, i.e. when form is *
                        # in which case the line she unchanged
                        except sre_constants.error:
                            updatedLine = line
                    else:
                        # line should be unchanged in case of no matches
                        updatedLine = line
                # write updated line to output
                outp.write(updatedLine)
            # lines without any tagged nouns should be copied as is
            else:
                outp.write(line)
        # close input and outp files
        inp.close()
        outp.close()
    # inform user of progress
    fileCounter += 1
    sys.stdout.write("\rProcessed file %d / %d " % (fileCounter,len(fileList)))
    sys.stdout.flush()

print "\nProcess complete!\n"
